// Copyright 1997 Ubilab, Union Bank of Switzerland. All rights reserved.
// This is part of the Geo system developed by Ubilab's DistrOOSE group.
// Check out: http://www.ubs.com/ubilab/Projects/se/se.html. GLAB-REH-97.

package Geo.Framework.AnyImpl;

import java.io.*;
import java.util.*;
import Geo.Framework.Any.*;



public abstract class AnyReaderDefImpl implements AnyReader {

	// class variables
	
	private static Vector readers;
	
	
	// static initialization
	
	static {
		readers = new Vector();
	}
	
	
	// instance variables
	
	protected StringTokenizer tokenizer;
	protected Stack pushbackStack;
	protected AnyContext context;
	protected Vector anyBuffer;
	protected Vector dummies;


	// class methods
	
	public static Any read(String in, AnyContext context) {
		StringTokenizer tokenizer = new StringTokenizer(in);
		String type = getInputType(tokenizer);	
		for (int i=0, s=readers.size(); i<s; i++) {
			if ( ((AnyReader) readers.elementAt(i)).getType().equals(type) ) {
				AnyReader anyReader = (AnyReader) readers.elementAt(i);
				return anyReader.readAny(in, context);
			}
		}
		throw new AnyReaderException("No reader found for type " + type);
	}
	

	// constructors
	
	protected AnyReaderDefImpl() {
		pushbackStack = new Stack();
		anyBuffer = new Vector();
		dummies = new Vector();
	}
	
	
	// accessing

	public abstract String getType();
	
	
	


public Any readAny(String in, AnyContext context) {
	this.context = context; 
	// set returnTokens to be able to handle spaces in strings
	this.tokenizer = new StringTokenizer(in, " \t\n\r", true);
	String type = getInputType(tokenizer);
	if ( !canRead(type) ) 
		throw new AnyReaderException(this, type);
	Any newAny = read();
	for (int i=0, s=dummies.size(); i<s; i++) 
		((AnyDummyHandle)dummies.elementAt(i)).replaceDummy(anyBuffer);
	anyBuffer.removeAllElements();
	return newAny;
}
protected String readNextToken() {
	if ( !pushbackStack.empty() )
		return (String) pushbackStack.pop();
	else {
		try {
			String token;
			while ( (token = tokenizer.nextToken()).equals(" ") ) {}
			return token; }
		catch (NoSuchElementException e) {
			throw new AnyReaderException(this);
		}
	}
}
private static String readNextToken(StringTokenizer tokenizer) {
	try {
		return tokenizer.nextToken(); }
	catch (NoSuchElementException e) {
		throw new AnyReaderException(); 
	}
}
protected String readNextTokenWithSpaces() {
	String token;
	StringBuffer b = new StringBuffer();
	try {
		token = tokenizer.nextToken();
		b.append(token); }
	catch (NoSuchElementException e) {
		throw new AnyReaderException(this);
	}
	while ( tokenizer.hasMoreTokens() && token.equals(" ") ) {
		token = tokenizer.nextToken();
		b.append(token); }
	return b.toString();  
}
public static void register(AnyReader reader) { 
		for (int i=0, s=readers.size(); i<s; i++) {
			if ( ((AnyReader) readers.elementAt(i)).getType().equals(reader.getType()))
				return;
		}
		readers.addElement(reader);
	}
	
	// private / protected
	
	protected boolean canRead(String type) {
		return getType().equals(type);
	}
	
	private static String getInputType(StringTokenizer tokenizer) {
		// input has to start with: //Any:
		String header = readNextToken(tokenizer);
		if ( !header.startsWith("//Any:") || header.length() < 7 ) 
			throw new AnyReaderException();
		return header.substring(6);
	}
	
	protected void pushBack(String token) {
		pushbackStack.push(token);
	}
	
	protected abstract Any read();
	
}
