// Copyright 1997 Ubilab, Union Bank of Switzerland. All rights reserved.
// This is part of the Geo system developed by Ubilab's DistrOOSE group.
// Check out: http://www.ubs.com/ubilab/Projects/se/se.html. GLAB-REH-97.

package Geo.Framework.AnyImpl;

import java.io.*;
import java.util.*;
import Geo.Framework.Any.*;



public class FastAnyReaderImpl extends AnyReaderDefImpl {

	// constructors
	public FastAnyReaderImpl() {
	}
	
	
	// accessing
	
	public String getType() {
		return "FastAnyIO";
	}	
	
	
	// protected - reading

	protected Any read() {
		Any newAny;
		String token = readNextToken();

		switch ( token.charAt(0) ) {
			case 'N': 
				newAny = AnySap.nullValue();
				break; 
			case '[': 
				newAny = readAnyArray();
				break; 
			case '{':
				newAny = readAnyDictionary();
				break;
			case '<':
				newAny = readAnyFrame();
				break;
			case '#':
				newAny = readAnyFrameDescriptor();
				break;
			case '!':
				newAny = readAnySlotDescriptor();
				break;
			case 'T':
				newAny = AnySap.trueValue();
				break;
			case 'F':
				newAny = AnySap.falseValue();
				break;
			case 'D': 
				newAny = readAnyDouble(token);
				break;
			case '0' : case '1' : case '2' : case '3' : case '4' : case '5' :
					   case '6' : case '7' : case '8' : case '9' : case '-' : 
				newAny = readAnyInteger(token);
				break;
			case '@' :
				newAny = readAnyString(token);
				break;
			case '*' :	
				try {
					int index = Integer.valueOf(token.substring(1)).intValue();
					Object element;
					if (index >= anyBuffer.size() 
						|| (element = anyBuffer.elementAt(index)) == null )
							return newAny = new AnyDummy(index);
					else
						newAny = (Any) element; }
				catch ( NumberFormatException e) {
					throw new AnyReaderException(this);
				}			
				break;
			default :
				throw new AnyReaderException(this);
		}
	
		String readAheadToken = readNextToken();
		if ( readAheadToken.startsWith("&") ) {
			int index = Integer.valueOf(readAheadToken.substring(1)).intValue();
			if ( anyBuffer.size() <= index )
				anyBuffer.setSize(index+1);
			anyBuffer.setElementAt(newAny, index); }
		else
			pushBack(readAheadToken);
		
		return newAny;
	}
	
	protected Any readAnyArray() {
		int size = read().asInteger();
		AnyArray anAnyArray = new AnyArrayImpl();
		for (int i=0; i<size; i++) {
			Any newAny = read();
			if (newAny instanceof AnyDummy)
				dummies.addElement(new AnyArrayDummyHandle(anAnyArray, i, ((AnyDummy)newAny).index));
			anAnyArray.append(newAny);
		}
		readNextToken(); // ignore ]
		return anAnyArray;
	}
	
	protected Any readAnyDictionary() {
		AnyDictionary anAnyDictionary = new AnyDictionaryImpl();
		try {
			AnyArray keys = read().asAnyArray();
			AnyArray values = read().asAnyArray();
			for (int i=0, s=keys.size(); i<s; i++) {
				AnyString key = keys.at(i).asAnyString();
				Any value = values.at(i);
				if (value instanceof AnyDummy)
					dummies.addElement(new AnyDictionaryDummyHandle(anAnyDictionary, key, ((AnyDummy)value).index));
				anAnyDictionary.setSlot(key, value);
			}
			readNextToken(); } // ignore }
		catch (AnyClassCastException e) {
			throw new AnyReaderException(this);
		}
		return anAnyDictionary;
	}
	
	protected Any readAnyDouble(String token) {
		try { 
			double d = Double.valueOf(token.substring(1)).doubleValue(); 
			return AnySap.doubleValue(d);}
		catch ( NumberFormatException e) {
			throw new AnyReaderException(this); 
		}
	}
	
	protected Any readAnyFrame() {
		try {
			AnyString name = read().asAnyString();
			AnyFrameDescriptor descriptor = context.getSymbolTable().at(name);
			if (descriptor==null) 
				throw new AnyReaderException("FrameDescriptor " + name + " not found in Symboltable");
			AnyArray slotValues = read().asAnyArray();
			AnyFrame frame = new AnyFrameImpl( context, name, slotValues);
			for (int i=0, s=slotValues.size(); i<s; i++) {
				Any newAny = slotValues.at(i);
				if (newAny instanceof AnyDummy) {
					AnyString slotName = descriptor.allSlotNames().at(i).asAnyString();
					dummies.addElement(new AnyFrameDummyHandle(frame, slotName, ((AnyDummy)newAny).index));
				}
			}
			readNextToken(); // ignore >
			return frame;  }
		catch (AnyClassCastException e) {
			throw new AnyReaderException(this);
		}
	}
	
	protected AnyFrameDescriptor readAnyFrameDescriptor() {
		try {
			AnyString name = read().asAnyString();
			AnyFrameDescriptor descriptor = new AnyFrameDescriptorImpl(context, name);
			Any superFrameDescriptor = read();
			if (superFrameDescriptor.notEqual(null))
				descriptor.setSuperFrameDescriptor(superFrameDescriptor.asAnyFrameDescriptor());
			descriptor.setContents(read().asAnyDictionary());
			readNextToken(); // ignore $
			return descriptor; }
		catch (AnyClassCastException e) {
			throw new AnyReaderException(this);
		}
	}
	
	protected Any readAnyInteger(String token) {
		try { 
			int i = Integer.valueOf(token).intValue(); 
			return AnySap.integerValue(i); }
		catch ( NumberFormatException e) {
			throw new AnyReaderException(this);
		}
	}
	
	protected Any readAnySlotDescriptor() {
		try {
			String flagsString = readNextToken();
			AnyString typeName = read().asAnyString();
			AnySlotDescriptor descriptor = new AnySlotDescriptorImpl(typeName);
			setFlags(descriptor, flagsString);	
			readNextToken(); // ignore $
			return descriptor; }
		catch (AnyClassCastException e) {
				throw new AnyReaderException(this);
		}
	}
	
	protected Any readAnyString(String token) {
		String s;
		int l = token.length();
		if ( !token.equals("@") && containsEndOfString(token) ) {
			// whole string in token
			s = token.substring(1,l-1); }
		else {
			StringBuffer b = new StringBuffer();
			if ( l>1 ) 
				b.append(token.substring(1));
			String nextPart = readNextTokenWithSpaces();
			while ( !containsEndOfString(nextPart) ) {
				b.append(nextPart);
				nextPart = readNextTokenWithSpaces();
			}	
			b.append( nextPart.substring(0,nextPart.length()-1) );
			s = b.toString();
		}
		s = removeDoubleDelimiter(s);
		return AnySap.stringValue(s);
	}
	

	// private - misc
	
	private boolean containsEndOfString(String token) {
		int n = 0;
		for (int i=token.length()-1; i>=0 && token.charAt(i) == '@'; i--)
			n++;
		return n % 2 == 1;
	}
	
	private void setFlags(AnySlotDescriptor aSlotDescriptor, String flagsString) {
		int flags;
		try { flags = Integer.valueOf(flagsString).intValue(); }
		catch (NumberFormatException e) {
			throw new AnyReaderException(this);
		}
		if ( (flags & FastAnyWriterImpl.SLOT_FLAG_SINGLE_VALUE) != 0 )
			aSlotDescriptor.setSingleValue();
		if ( (flags & FastAnyWriterImpl.SLOT_FLAG_MAY_NOT_REMOVE) != 0 )
			aSlotDescriptor.setMayNotRemove();
		if ( (flags & FastAnyWriterImpl.SLOT_FLAG_MUST_HAVE) != 0 )
			aSlotDescriptor.setMustHave();
		if ( (flags & FastAnyWriterImpl.SLOT_FLAG_NOT_EDITABLE) != 0 )
			aSlotDescriptor.setNotEditable();	
	}
	
	private String removeDoubleDelimiter(String s) {
		boolean firstWritten = false;
		StringTokenizer t = new StringTokenizer(s, "@", true);
		StringBuffer b = new StringBuffer();
		while ( t.hasMoreTokens() ) {
			String token = t.nextToken();
			if ( token.equals("@") ) {
				if ( !firstWritten ) {
					b.append(token);
					firstWritten = true; }
				else
					firstWritten = false; }
			else
				b.append(token);
		}	
		return b.toString();
	}
	
}
