// Copyright 1997 Ubilab, Union Bank of Switzerland. All rights reserved.
// This is part of the Geo system developed by Ubilab's DistrOOSE group.
// Check out: http://www.ubs.com/ubilab/Projects/se/se.html. GLAB-REH-97.

package Geo.Framework.AnyImpl;

import Geo.Framework.Any.*;



public class FastAnyWriterImpl extends AnyWriterDefImpl {

	// constants
	
	public static int SLOT_FLAG_MASK = 0x00ffffff;
	public static int SLOT_FLAG_SINGLE_VALUE = 1 << 0;
	public static int SLOT_FLAG_MAY_NOT_REMOVE = 1 << 1;
	public static int SLOT_FLAG_MUST_HAVE = 1 << 2;
	public static int SLOT_FLAG_NOT_EDITABLE = 1 << 3;


	// constructors
	
	public FastAnyWriterImpl() {
	}
	
	
	// accessing
	
	public String getType() {
		return "FastAnyIO";
	}
	
	
	// protected - writing
	
	protected void write(Any anAny) {
		// only object value types are buffered
		if (anAny != null && anAny.isObjectType()) {
			int referencedId = this.anyBufferIndexOf(anAny);
			if ( referencedId == -1 ) {
				int id = this.buffer(anAny); // it was not in buffer
				dispatchWrite(anAny);				
				addToBuffer(" &" + id); }
			else 
				addToBuffer(" *" + referencedId); }// it was in buffer
		else 
			dispatchWrite(anAny);
	}
	
	protected void writeAnyArray(AnyArray anAnyArray) {
		addToBuffer(" [ " + anAnyArray.size());
		for (int i=0, s=anAnyArray.size(); i<s; i++) 
			write(anAnyArray.at(i)); 
		addToBuffer(" ]");
	}
	
	protected void writeAnyBoolean(AnyBoolean anAnyBoolean) {
		if (anAnyBoolean.asBoolean()) 
			addToBuffer(" T");
		else
			addToBuffer(" F");
	}
	
	protected void writeAnyDictionary(AnyDictionary anAnyDictionary) {
		addToBuffer(" {");
		write(anAnyDictionary.keys());
		write(anAnyDictionary.values());
		addToBuffer(" }");
	}
	
	protected void writeAnyDouble(AnyDouble anAnyDouble) {
		addToBuffer(" D" + anAnyDouble);
	}
	protected void writeAnyFrame(AnyFrame anAnyFrame) {
		addToBuffer(" <");
		write(anAnyFrame.getFrameDescriptor().getName());
		write(anAnyFrame.slotValues());
		addToBuffer(" >");
	}
	
	protected void writeAnyFrameDescriptor(AnyFrameDescriptor anAnyFrameDescriptor) {
		addToBuffer(" #");
		write(anAnyFrameDescriptor.getName());
		write(anAnyFrameDescriptor.getSuperFrameDescriptor());
		write(anAnyFrameDescriptor.getContents());
		addToBuffer(" $");
	}
	
	protected void writeAnyInteger(AnyInteger anAnyInteger) {
		addToBuffer(" " + anAnyInteger);
	}
	
	protected void writeAnySlotDescriptor(AnySlotDescriptor anAnySlotDescriptor) {
		addToBuffer(" !");
		writeAnySlotDescriptorFlags(anAnySlotDescriptor);
		write(anAnySlotDescriptor.getTypeName());
		addToBuffer(" $");
	}
	
	private void writeAnySlotDescriptorFlags(AnySlotDescriptor anAnySlotDescriptor) {
		int flags = 0;
		if ( anAnySlotDescriptor.isSingleValue() )
			flags = flags | SLOT_FLAG_SINGLE_VALUE & SLOT_FLAG_MASK;
		if ( anAnySlotDescriptor.isMayNotRemove() )
			flags = flags | SLOT_FLAG_MAY_NOT_REMOVE & SLOT_FLAG_MASK;
		if ( anAnySlotDescriptor.isMustHave() )
			flags = flags | SLOT_FLAG_MUST_HAVE & SLOT_FLAG_MASK;	
		if ( anAnySlotDescriptor.isNotEditable() )
			flags = flags | SLOT_FLAG_NOT_EDITABLE & SLOT_FLAG_MASK;
		addToBuffer(" " + String.valueOf(flags));
	}
	
	protected void writeAnyString(AnyString anAnyString) {
		StringBuffer tmp = new StringBuffer();
		addToBuffer(" @");
		for (int i=0, s=anAnyString.asString().length(); i<s; i++) {
			char c = anAnyString.asString().charAt(i);
			if (c == '@' )
				tmp.append("@");
			tmp.append(c);
		}
		addToBuffer(tmp.toString());
		addToBuffer("@");
	}
	
	protected void writeNull() {
		addToBuffer(" N");
	}
	
}
